unit msg_child;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes,
  System.Variants,
  System.Messaging, System.SyncObjs, System.Generics.Collections,
  FMX.Types, FMX.Controls, FMX.Forms, FMX.Graphics, FMX.Dialogs, FMX.StdCtrls,
  FMX.Controls.Presentation, FMX.ScrollBox, FMX.Memo, FMX.Layouts, FMX.ListBox,
  msg_structures, System.Diagnostics;

type
  TProc = reference to procedure(const AText: string);

type
  TfrmChild = class(TForm)
    btnThreadAction: TButton;
    lblMessageText: TLabel;
    btnMessage2: TButton;
    lbxMessages: TListBox;
    memMessages: TMemo;
    chkDelay: TCheckBox;
    btnSendToThread: TButton;
    tmrMsgSend: TTimer;
    btnToThread4: TButton;
    tmrMsgThread4: TTimer;
    procedure btnMessage2Click(Sender: TObject);
    procedure btnThreadActionClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btnSendToThreadClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure tmrMsgSendTimer(Sender: TObject);
    procedure btnToThread4Click(Sender: TObject);
    procedure tmrMsgThread4Timer(Sender: TObject);
  private
    {Private declarations}
    MessageManager: TMessageManager;
    FMsgRegId: Integer;
    FMsgProc: TProc;
    FItemIndex: Integer;
    FMsgRegId2: Integer;
    FMsgQueue: TThreadedQueue<Integer>;
    FMsgQueue4: TThreadedQueue<Integer>;
    FMsgRegId4: Integer;

    procedure _DisplayMessage(const AText: string);
    procedure _SendListItem(AValue: Integer);
    procedure _SendListItem4(AValue: Integer);
  public
    {Public declarations}
  end;

var
  frmChild: TfrmChild;

implementation

//uses
//  tbsLogger;

{$R *.fmx}

procedure TfrmChild.btnMessage2Click(Sender: TObject);
var
  MessageManager: TMessageManager;
  lMessage: TMessage;
begin
  MessageManager := TMessageManager.DefaultManager;
  lMessage := TMessage<Integer>.Create(2);
  MessageManager.SendMessage(Sender, lMessage, True);
end;

procedure TfrmChild.btnSendToThreadClick(Sender: TObject);
begin
  btnSendToThread.Enabled := False;
  try
    lbxMessages.ItemIndex := FItemIndex;
//    TtbsGCLogger.Send('3 send message start', FItemIndex);
    inc(FItemIndex);
    _SendListItem(FItemIndex);
    if FItemIndex = lbxMessages.Count then
      FItemIndex := 0;
  finally
    btnSendToThread.Enabled := True;
  end;
end;

procedure TfrmChild.btnThreadActionClick(Sender: TObject);
var
  lStopWatch: TStopwatch;
begin
//  TtbsGCLogger.Send('send message start', 3);
  lStopWatch := TStopWatch.StartNew;
  TThread.CreateAnonymousThread(

    procedure()
    var
      MessageManager: TMessageManager;
      lMessage: TMessage;
    begin
//      TtbsGCLogger.Send('send message thread', 3);

      MessageManager := TMessageManager.DefaultManager;
      lMessage := TMessage<Integer>.Create(3);
      MessageManager.SendMessage(Sender, lMessage, True);
//      TtbsGCLogger.Send('sent message', 3);

    end).Start;
//  TtbsGCLogger.Send('btn click elapsed', lStopWatch.Elapsed);
end;

procedure TfrmChild.btnToThread4Click(Sender: TObject);
begin
  btnToThread4.Enabled := False;
  try
    lbxMessages.ItemIndex := FItemIndex;
    inc(FItemIndex);
    _SendListItem4(FItemIndex);
    if FItemIndex = lbxMessages.Count then
      FItemIndex := 0;
  finally
    btnToThread4.Enabled := True;
  end;
end;

procedure TfrmChild.FormCreate(Sender: TObject);
begin
  MessageManager := TMessageManager.DefaultManager;

  FMsgProc := procedure(const AText: string)
    begin
      lblMessageText.Text := AText;
      memMessages.Lines.Add(AText);
      if memMessages.Lines.Count > 15 then
        memMessages.Lines.Delete(0);
    end;

  FMsgQueue := TThreadedQueue<Integer>.Create(10, 1, 1);
  FMsgQueue4 := TThreadedQueue<Integer>.Create(10, 1, 1);

  // this is for thread 1, thread 2
  // thread 1 sends from an anon thread, thread 2 just sends
  FMsgRegId := MessageManager.SubscribeToMessage(TMessage<UnicodeString>,
    procedure(const Sender: TObject; const Msg: TMessage)
    var
      sMsg: string;
    begin
      sMsg := (Msg as TMessage<UnicodeString>).Value;
      _DisplayMessage(sMsg);
    end);

  // message for integers from thread 3

  FMsgRegId2 := MessageManager.SubscribeToMessage(TMessage<TFromThreadInteger>,
    procedure(const Sender: TObject; const Msg: TMessage)
    var
      iValue: Integer;
      lMsg: TFromThreadInteger;
    begin
      if Sender <> Self then
      begin
      iValue := (Msg as TMessage<TFromThreadInteger>).Value;
      FMsgQueue.PushItem(iValue);
//      TtbsGCLogger.Send('child from thread received ', iValue);
      TThread.Synchronize(TThread.CurrentThread,
        procedure()
        begin
          tmrMsgSend.Enabled := True;
        end);
      end;
    end);

  // message for integers from thread 4

  FMsgRegId4 := MessageManager.SubscribeToMessage(TMessage<TMyFromThreadMessage>,
    procedure(const Sender: TObject; const Msg: TMessage)
    var
      iValue: Integer;
      lMsgValue: TMyFromThreadMessage;
    begin
      lMsgValue := (Msg as TMessage<TMyFromThreadMessage>).Value;
      iValue := lMsgValue.Value;
      FMsgQueue4.PushItem(iValue);
//      TtbsGCLogger.Send('child from thread 4 received ', iValue);
      TThread.Synchronize(TThread.CurrentThread,
        procedure()
        begin
          tmrMsgThread4.Enabled := True;
        end);
    end);

end;

procedure TfrmChild.FormDestroy(Sender: TObject);
begin
  MessageManager.Unsubscribe(TMessage<UnicodeString>, FMsgRegId);
  MessageManager.Unsubscribe(TMessage<TFromThreadInteger>,FMsgRegId2);
  MessageManager.Unsubscribe(TMessage<TMyFromThreadMessage>, FMsgRegId4);
  FMsgQueue.Free;
  FMsgQueue4.Free;
end;

procedure TfrmChild.tmrMsgThread4Timer(Sender: TObject);
var
  iValue: Integer;
  iPopResult: TWaitResult;
  iQueSize: Integer;
begin
  tmrMsgThread4.Enabled := False;
  iPopResult := FMsgQueue4.PopItem(iQueSize, iValue);
  if iPopResult = wrSignaled then
  begin
//    TtbsGCLogger.Send('4 value', iValue);
    if iValue = FItemIndex then
    begin
      inc(FItemIndex);
      lbxMessages.ItemIndex := iValue;
      if FItemIndex = 10 then
      begin
        FItemIndex := 0;
      end;
      _SendListItem4(FItemIndex);
    end;
  end;

end;

procedure TfrmChild.tmrMsgSendTimer(Sender: TObject);
var
  iValue: Integer;
  iPopResult: TWaitResult;
  iQueSize: Integer;
begin
  tmrMsgSend.Enabled := False;
  try
    iPopResult := FMsgQueue.PopItem(iQueSize, iValue);
//    TtbsGCLogger.Send('3 pop result', Integer(iPopResult));
//    TtbsGCLogger.Send('value', iValue);
    if iPopResult = wrSignaled then
    begin
//      TtbsGCLogger.Send('3 pop value', iValue);
      if iValue = FItemIndex then
      begin
        inc(FItemIndex);
        lbxMessages.ItemIndex := iValue;
        if FItemIndex = 10 then
        begin
          FItemIndex := 0;
        end;
        _SendListItem(FItemIndex);
      end;
    end;
  finally
    // tmrMsgSend.Enabled := True;
  end;
end;

procedure TfrmChild._DisplayMessage(const AText: string);
begin
      TThread.Synchronize(TThread.CurrentThread,
        procedure()
        begin
          if chkDelay.IsChecked then
            Sleep(500);

          FMsgProc(AText);
//          TtbsGCLogger.Send('child display', AText);
        end);

end;

procedure TfrmChild._SendListItem(AValue: Integer);
var
  lMessage: TMessageBase;
  iValue: Integer;
begin
  iValue := FItemIndex;
  lMessage := TMessage<TToThreadInteger>.Create(iValue);
  MessageManager.SendMessage(Self, lMessage);
//  TtbsGCLogger.Send('sent list item', iValue);

end;

procedure TfrmChild._SendListItem4(AValue: Integer);
var
  lMessage: TMessageBase;
  iValue: Integer;
  lMsgValue: TMyToThreadMessage;
begin
  iValue := FItemIndex;
  lMsgValue := TMyToThreadMessage.Create(iValue);
  lMessage := TMessage<TMyToThreadMessage>.Create(lMsgValue);
  MessageManager.SendMessage(Self, lMessage, True);
//  TtbsGCLogger.Send('4 sent list item', iValue);

end;

end.
